unit XMS;
{****************************************************************************
** Unit allowing the use of extended memory in Turbo Pascal programs.      **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************}

interface
  function XMSInit : Boolean;
  function XMSLargestAvailable : Word;
  function XMSTotalAvailable : Word;
  procedure XMSGetMem (var Handle : Word; Size : Word);
  procedure XMSFreeMem (Handle : Word);
  procedure XMSCopy (SHandle : Word; SOfs : LongInt; THandle : Word; TOfs, Amount : LongInt);
  procedure XMSLock (Handle : Word; var LinearAddress : LongInt);
  procedure XMSUnLock (Handle : Word);

implementation

  uses Dos;

  {Structure used for copying data}
  type EMMS = record
    Amount : LongInt;
    SHandle : Word;
    SOfs : LongInt;
    THandle : Word;
    TOfs : LongInt;
  end;

  {Structure used in calling the XMS manager}
  type XMSRegs = record
    AX, BX, DX, SI, DS : Word;
  end;

  var
    {Pointer to the XMS manager}
    XMSPtr : pointer;
    {Registers}
    xAX, xBX, xDX, xSI, xDS : Word;

{****************************************************************************
** Calls the extended memory manager.                                      **
****************************************************************************}
procedure XMSCall (FunctionNumber : Byte; var XRegs : XMSRegs); assembler;
asm
  mov ah, FunctionNumber {Retrieve function number}

  mov cx, ds             {Copy DS into CX}
  push ds                {And store it}

  lds di, [bp+4]         {Retrieve pointer to register values}
  mov bx, [di+2]         {Retrieve BX}
  mov dx, [di+4]         {Retrieve DX}
  cmp ah, 11             {If it is not a move operation, we don't need to...}
  jne @NoMoveOperation

  {set up the structure used for data copying}
  mov si, [di+6]
  mov ds, [di+8]

@NoMoveOperation:
  mov es, cx             {The pointer XMSPtr is stored in the original DS, which was copied into CX}
  call es:[XMSPtr]       {The actual call is made here}
  mov cx, ds             {The new value of DS must be stored}
  lds di, [bp+4]         {Get the pointer to the registers again}
  mov [di], ax           {Return registers}
  mov [di+2], bx
  mov [di+4], dx
  mov [di+6], si
  mov [di+8], cx
  pop ds                 {Restore DS}
end;

{****************************************************************************
** Checks to see if an extended memory manager has been installed.         **
** returns true if XMS manager found, false if not.                        **
****************************************************************************}
function XMSInit : Boolean;
var
  Regs : Registers;

begin
  {Multiplexer code 4300h - HIMEM.SYS installation status}
  Regs.AX := $4300;
  Intr($2F, Regs);
  {Returns AL = 80h if installed}
  if (Regs.AL = $80) then begin
    {Multiplexer code 4310h - Store address for calling XMS functions}
    Regs.AX := $4310;
    intr($2F, Regs);
    XMSPtr := Ptr(Regs.ES, Regs.BX);
    XMSInit := true;
  end else
   XMSInit := false;
end;

{****************************************************************************
** Returns the size in K of the largest XMS block that can be allocated    **
****************************************************************************}
function XMSLargestAvailable : Word;
var
  Xr : XMSRegs;

begin
  {Set up registers to call XMS manager}
  Xr.AX := 0;
  Xr.BX := 0;
  Xr.DX := 0;
  Xr.SI := 0;
  Xr.DS := 0;
  {Call XMS function 8 - query free extended memory}
  XMSCall (8, Xr);
  {Return size of largest available clock}
  XMSLargestAvailable := Xr.AX;
end;

{****************************************************************************
** Returns the total amount of free XMS in K                               **
****************************************************************************}
function XMSTotalAvailable : Word;
var
  Xr : XMSRegs;

begin
  {Set up registers to call XMS manager}
  Xr.AX := 0;
  Xr.BX := 0;
  Xr.DX := 0;
  Xr.SI := 0;
  Xr.DS := 0;
  {Call XMS function 8 - query free extended memory}
  XMSCall (8, Xr);
  {Return total amount available}
  XMSTotalAvailable := Xr.DX;
end;

{****************************************************************************
** Allocates an extended memory block of Size K-bytes and returns a handle **
****************************************************************************}
procedure XMSGetMem (var Handle : Word; Size : Word);
var
  Xr : XMSRegs;

begin
  {Set up registers to call XMS manager}
  Xr.AX := 0;
  Xr.BX := 0;
  Xr.DX := Size;
  Xr.SI := 0;
  Xr.DS := 0;
  {Call XMS function 9 - allocate extended memory block}
  XMSCall (9, Xr);

  {Error if AX is 0 and BX > 80h}
  if (Xr.AX = 0) and (Xr.BX >= 128) then
    {If so, return 0}
    Handle := 0
  else
    {Otherwise, return handle}
    Handle := Xr.DX;
end;

{****************************************************************************
** Frees an extended memory block allocated by XMSGetMem                   **
****************************************************************************}
procedure XMSFreeMem (Handle : Word);
var
  Xr : XMSRegs;

begin
  {Set up registers to call XMS manager}
  Xr.AX := 0;
  Xr.BX := 0;
  Xr.DX := Handle;
  Xr.SI := 0;
  Xr.DS := 0;
  {Call XMS function 10 - free extended memory block}
  XMSCall (10, Xr);
end;

{****************************************************************************
** Copies "amount" bytes of data from a memory block SHandle starting at   **
** offset SOfs within that block to another memory block THandle at offset **
** TOfs within that block. To copy to/from conventional memory, specify    **
** handle number 0 and use a SEGM:OFFS pointer for offset values.          **
*****************************************************************************
** NOTE: This function can only copy EVEN numbers of bytes. A single byte  **
** copy is therefore not possible.                                         **
****************************************************************************}
procedure XMSCopy (SHandle : Word; SOfs : LongInt; THandle : Word; TOfs, Amount : LongInt);
var
  Mi : EMMS;
  Xr : XMSRegs;

begin
  {Set up structure to pass to the XMS manager}
  Mi.Amount := Amount;
  Mi.SHandle := SHandle;
  Mi.SOfs := SOfs;
  Mi.THandle := THandle;
  Mi.TOfs := TOfs;
  Xr.SI := Ofs (Mi);
  Xr.DS := Seg (Mi);
  {Call XMS function 11 - copy data}
  XMSCall (11, Xr);
end;

{****************************************************************************
** This locks the position of the EMB in XMS memory. The XMS manager often **
** moves EMBs around to allow more efficient allocation.                   **
****************************************************************************}
procedure XMSLock (Handle : Word; var LinearAddress : LongInt);
var
  Xr : XMSRegs;

begin
  {Set up registers to call XMS manager}
  Xr.AX := 0;
  Xr.BX := 0;
  Xr.DX := Handle;
  Xr.SI := 0;
  Xr.DS := 0;
  {Call XMS function 12 - lock extended memory block}
  XMSCall (12, Xr);
  {Return linear address of EMB}
  LinearAddress := Xr.DX shl 16 or Xr.BX;
end;

{****************************************************************************
** This unlocks the position of the EMB in XMS memory.                     **
****************************************************************************}
procedure XMSUnLock (Handle : Word);
var
  Xr : XMSRegs;

begin
  {Set up registers to call XMS manager}
  Xr.AX := 0;
  Xr.BX := 0;
  Xr.DX := Handle;
  Xr.SI := 0;
  Xr.DS := 0;
  {Call XMS function 13 - unlock extended memory block}
  XMSCall (13, Xr);
end;

begin
  if XMSInit then
    Writeln ('XMS routines installed.')
  else
    Writeln ('No XMS found.');
end.